/*
 * COPYRIGHT INFORMATION - DO NOT REMOVE
 * "Portions Copyright (c) 2000-2001 LinuxMagic Inc. All Rights Reserved.
 *
 * This file contains Original Code and/or Modifications of Original Code as
 * defined in and that are subject to the Free Source Code License Version
 * 1.0 (the 'License'). You may not use this file except in compliance with
 * the License. Please obtain a copy of the License at:
 *
 * http://www.linuxmagic.com/opensource/licensing/
 *
 * and read it before using this file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND LINUXMAGIC HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT. Please see
 * the License for the specific language governing rights and limitations
 * under the License."
 *
 * Please read the terms of this license carefully. By using or downloading
 * this software or file, you are accepting and agreeing to the terms of this
 * license with LinuxMagic Inc. If you are agreeing to this license on behalf
 * of a company, you represent that you are authorized to bind the company to
 * such a license. If you do not meet this criterion or you do not agree to
 * any of the terms of this license, do NOT download, distribute, use or alter
 * this software or file in any way.
 *
 * Author: Burton Samograd (burton@wizard.ca)
 * CVS Id: $Id: auth_db.c,v 1.6 2003/01/15 18:25:47 burton Exp $
 *
 * DO NOT MODIFY WITHOUT CONSULTING THE LICENSE
 */

#include <crypt.h>
#include <string.h>

#include <liblm.h>

#include "magic-smtpd.h"
#include "config_msd.h"
#include "config_qmail.h"
#include "smtp.h"
#include "auth.h"
#include "db.h"
#include "b64.h"

int msd_auth_init_db(void)
{
  int retval = 0;
  /* Nothing to do here, handled by db.c */
  LM_ASSERT(db != NULL);
  if(db == NULL)
    {
      /* Can't do auth when db is down */
      return -1;
    }

  return retval;
}

int msd_auth_cleanup_db(void)
{
  int retval = 0;
  /* Nothing to do here, handled by db.c */
  LM_ASSERT(db != NULL);
  if(db == NULL)
    {
      /* Can't do auth when db is down */
      return -1;
    }
  return retval;
}

/* Connect to database and authorize user */
int msd_do_auth_db(int type, lm_string_t username64)
{
  char query[1024];
  lm_string_t username = NULL;
  lm_string_t password = NULL;
  lm_string_t password64 = NULL; 
  int retval = 0;
  PGresult *result;
  char *escuser = NULL;
  char *escdomain = NULL;
	
  if(db == NULL)
    {
      /* Can't do auth when db is down */
      return LM_AUTH_ERROR_DB;
    }

  /* Only support LOGIN auth type for now */
  if(type != SMTP_AUTH_LOGIN)
    {
      SYSLOG((LOG_INFO, "Unsupported authorization type requested (not LOGIN)"));
      return LM_AUTH_UNSUPPORTED;
    }

  /* Print the Username: line in base64, read an input line and
     decode it to get the username */
  if(username64 == NULL)
    {
      printf("334 VXNlcm5hbWU6\r\n"); /* Username: */
      fflush(stdout);
      msd_timeout_read(0, &username64);
      LM_STRING_CHOP(username64);
      /* check for user cancel '*' */
      if(LM_STRING_LEN(username64) == 1 &&
	 LM_STRING_BUFFER(username64)[0] == '*')
	{
	  /* user canceled */
	  LM_STRING_FREE(username64);
	  return LM_AUTH_CANCEL;
	}
    }
  if(b64decode(username64, &username))
    {
      LM_ASSERT(username == NULL);
      LM_STRING_FREE(username64);
      return LM_AUTH_BAD_DATA;
    }
  else
    {
      LM_STRING_CHOP(username);
    }
  LM_STRING_FREE(username64);

  /* Print the Password: line in base64, read an input line and
     decode it to get the password */
  printf("334 UGFzc3dvcmQ6\r\n"); /* Password: */
  fflush(stdout);
  msd_timeout_read(0, &password64);
  LM_STRING_CHOP(password64);
  /* check for user cancel '*' */
  if(LM_STRING_LEN(password64) == 1 &&
     LM_STRING_BUFFER(password64)[0] == '*')
    {
      /* user canceled */
      LM_STRING_FREE(username);
      LM_STRING_FREE(password64);
      return LM_AUTH_CANCEL;
    }
  if(b64decode(password64, &password))
    {
      /* failure to decode */
      LM_ASSERT(password == NULL);
      LM_STRING_FREE(username);
      LM_STRING_FREE(password64);
      return LM_AUTH_BAD_DATA;
    }
  else
    {
      LM_STRING_CHOP(password);
    }
  LM_STRING_FREE(password64);

  /* 
   * call the init function now.
   */
  LM_ASSERT(msd_auth_init);
  if(msd_auth_init())
    {
      SYSLOG((LOG_ERR, "Could not open database connection for AUTH"));
      return LM_AUTH_ERROR_DB;
    }

  /* escape the data passed by the user for the query */
  msd_db_get_user_domain(&escuser, &escdomain, username);
  LM_STRING_FREE(username); /* don't need this anymore */
	
  /* TODO make this query string a config variable */
  snprintf(query, sizeof(query),
	   "select pw_passwd from vpopmail "
	   "where "
	   "pw_name='%s' "
	   "and "
	   "pw_domain='%s'",
	   escuser, escdomain);

  /* TODO check to make sure the db connection is open */
  retval = db_do_query(&result, query);
  if(retval < 0)
    {
      /* Didn't get any results so user is not in the database? */
#ifdef _LMDEBUG
      printf("%s\n", PQresStatus(PQresultStatus(result)));
      printf("%s\n", PQresultErrorMessage(result));
#endif
    }
  else 
    {
      int ntuples = PQntuples(result);
      int nfields = PQnfields(result);

      /* Got some results */
      if(ntuples == 1 && nfields == 1)
	{

	  lm_string_t retpwd = NULL;
	  char *tmppwd;
	  char *cryptpwd;

	  tmppwd = PQgetvalue(result, 0, 0);

	  if(tmppwd != NULL)
	    {
	      LM_STRING_NEW(retpwd, tmppwd);
	    }
	  else
	    {
	      /* Wierd error */
	      SYSLOG((LOG_ERR, "Bad data returned from database"));
	      retval = LM_AUTH_ERROR;
	    }

	  cryptpwd = crypt(LM_STRING_BUFFER(password), 
			   LM_STRING_BUFFER(retpwd));
	  if(cryptpwd != NULL && retval == 0)
	    {
	      if(!strncmp(cryptpwd, LM_STRING_BUFFER(retpwd), LM_STRING_LEN(retpwd)))
		{
		  /* Passwords match */
		  /* Kill the relayclient value and set remoteinfo to the username */
		  if(config_qmail.tcprelayclient)
		    LM_STRING_FREE(config_qmail.tcprelayclient);
		  LM_STRING_NEW(config_qmail.tcprelayclient, "");
		  LM_STRING_FREE(config_qmail.tcpremoteinfo);
		  LM_STRING_NEW(config_qmail.tcpremoteinfo, escuser);
		  SYSLOG((LOG_INFO, "AUTH success: %s@%s", 
			 escuser, escdomain));
		  retval = 0;
		}
	      else
		{
		  /* Passwords don't match */
		  SYSLOG((LOG_INFO, "AUTH failure: %s@%s", 
			 escuser, escdomain));
		  retval = LM_AUTH_FAILED;
		}
	    }
	  LM_STRING_FREE(retpwd);
	}
      else
	{
	  /* User not in database? */
	  SYSLOG((LOG_INFO, "AUTH failure user not in db: %s@%s", 
		 escuser, escdomain));
	  retval = LM_AUTH_NO_USER;
	}
    }

  LM_STRING_FREE(password);
  LM_FREE(escuser);
  LM_FREE(escdomain);

  return retval;
}
