/*
 * COPYRIGHT INFORMATION - DO NOT REMOVE
 * "Portions Copyright (c) 2000-2001 LinuxMagic Inc. All Rights Reserved.
 *
 * This file contains Original Code and/or Modifications of Original Code as
 * defined in and that are subject to the Free Source Code License Version
 * 1.0 (the 'License'). You may not use this file except in compliance with
 * the License. Please obtain a copy of the License at:
 *
 * http://www.linuxmagic.com/opensource/licensing/
 *
 * and read it before using this file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND LINUXMAGIC HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT. Please see
 * the License for the specific language governing rights and limitations
 * under the License."
 *
 * Please read the terms of this license carefully. By using or downloading
 * this software or file, you are accepting and agreeing to the terms of this
 * license with LinuxMagic Inc. If you are agreeing to this license on behalf
 * of a company, you represent that you are authorized to bind the company to
 * such a license. If you do not meet this criterion or you do not agree to
 * any of the terms of this license, do NOT download, distribute, use or alter
 * this software or file in any way.
 *
 * Author: Burton Samograd (burton@wizard.ca)
 * CVS Id: $Id: check_user_fs.c,v 1.9 2003/01/15 18:25:47 burton Exp $
 *
 * DO NOT MODIFY WITHOUT CONSULTING THE LICENSE
 */

/* C includes */
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include <errno.h>
#include <sys/stat.h>
#include <pwd.h>

#include <liblm.h>

#include "magic-smtpd.h"
#include "config_msd.h"
#include "config_qmail.h"
#include "check_user.h"
#include "cdb.h"
#include "smtp.h"

/* Module globals */
static int is_vdomain_addr;
static int matched_dash_default;

/* Constants strings allocated during init */
static lm_string_t dotqmail = NULL;
static lm_string_t dotmagicspam = NULL;
static lm_string_t dashdefault = NULL;
static lm_string_t slash = NULL;
static lm_string_t dash = NULL;
static lm_string_t bang = NULL;
static lm_string_t alias = NULL;
static lm_string_t qmailalias = NULL;
static lm_string_t qmailaliashome = NULL;
static lm_string_t spam_users_dir = NULL;
static FILE *cdbusers = NULL;

/* Module static functions */
static int msd_check_vdomain_user(lm_string_t user, lm_string_t domain);
static int msd_build_vdomain_user(lm_string_t *puser, lm_string_t hostname);
static int msd_check_rcpt_system_user(lm_string_t user, lm_string_t domain, lm_string_t* psysuser, lm_string_t* pdotspam);
static int msd_check_rcpt_alias(lm_string_t user, lm_string_t domain, lm_string_t* psysuser, lm_string_t *pdotspam);
static int msd_check_rcpt_qmail_users(lm_string_t user, lm_string_t domain, lm_string_t* psysuser, lm_string_t* pdotspam);
static int msd_user_spam_enabled(lm_string_t user);

/* Allocate strings and open the cdb file */
int msd_check_user_init_fs(void)
{
	int retval;
	lm_string_t qmaildash = NULL;
	lm_string_t spam_users = NULL;
	struct passwd *pw;

	retval = 0;
	LM_STRING_NEW(dashdefault, "-default");
	retval |= (dashdefault == NULL);
	LM_STRING_NEW(qmaildash, ".qmail-");
	retval |= (qmaildash == NULL);
	LM_STRING_NEW(dotqmail, ".qmail");
	retval |= (dotqmail == NULL);
	LM_STRING_NEW(dotmagicspam, ".magicspam");
	retval |= (dotmagicspam == NULL);
	LM_STRING_NEW(slash, "/");
	retval |= (slash == NULL);
	LM_STRING_NEW(dash, "-");
	retval |= (dash == NULL);
	LM_STRING_NEW(bang, "!");
	retval |= (bang == NULL);
	LM_STRING_NEW(alias, "alias");
	retval |= (alias == NULL);

	/* Open the cdb users file if present */
	cdbusers = fopen("/var/qmail/users/cdb", "rb");
	/* Don't check to see if the file opened sucessfully, since the cdb
	   functions check for a null file pointer when they are called.  This saves
	   from having to write code before every cdb lookup to check for an open
	   file */

	/* Search for qmail alias directory in the cdb file first */
	if(cdb_lookup(cdbusers, "!alias", 7, &qmailaliashome))
	{
		/* CDB lookup failed, check the password file */
		/* Build the prefix for the qmail alias directory entries */
		pw = getpwnam("alias");
		if(pw == NULL)
		{
			/* Not our problem if qmail isn't setup properly */
			LM_ASSERT(qmailaliashome == NULL);
			LM_STRING_NEW(qmailaliashome, "");
		}
		else
		{
			/* Build the string ~alias/.qmail- for checking aliases later */
			LM_ASSERT(qmailaliashome == NULL);
			LM_STRING_NEW(qmailaliashome, pw->pw_dir);
		}
	}
	/* Build the string ~alias/.qmail- which we will use later
	   for verifying users */
	if(LM_STRING_LEN(qmailaliashome) > 1)
	{
		if(LM_STRING_DUP(qmailalias, qmailaliashome) ||
				LM_STRING_CAT(qmailalias, slash) ||
				LM_STRING_CAT(qmailalias, qmaildash))
			retval |= qmailalias == NULL;
	}
	LM_STRING_FREE(qmaildash);

	/* Build the spam users control dir */
	if(LM_STRING_NEW(spam_users, "spam_users/") ||
			LM_STRING_DUP(spam_users_dir, config.controldir) ||
			LM_STRING_CAT(spam_users_dir, spam_users))
		retval |= (spam_users_dir == NULL);
	LM_STRING_FREE(spam_users);

	return retval;
}

int msd_check_user_cleanup_fs(void)
{
	/* These will always be non null at this point, since an allocation
	   failure would have occured earlier */
	LM_STRING_FREE(dashdefault);
	LM_STRING_FREE(qmailalias);
	LM_STRING_FREE(qmailaliashome);
	LM_STRING_FREE(spam_users_dir);
	LM_STRING_FREE(dotqmail);
	LM_STRING_FREE(dotmagicspam);
	LM_STRING_FREE(slash);
	LM_STRING_FREE(dash);
	LM_STRING_FREE(bang);
	LM_STRING_FREE(alias);
	if(cdbusers != NULL)
		fclose(cdbusers);
	return 0;
}

/* Check if the user exists by performing the following actions:
 * see doc/lmsmtpd-check-user.txt
 */
int msd_check_rcpt_user_fs(lm_string_t* paddr, lm_string_t* pdotspam)
{
	lm_string_t addr = NULL;
	lm_string_t user = NULL;
	lm_string_t sysuser = NULL;
	lm_string_t domain = NULL;
	lm_string_t dotspam = NULL;
	int index;
	int retval;
	int is_vdomain_user, is_system_user, is_alias;
	int retcode;

	LM_ASSERT(paddr);
	LM_STRING_ASSERT(*paddr);
	addr = *paddr;

	/* Reset state variables for user type */
	is_vdomain_addr = 0;
	is_system_user = 0;

	/* This is already checked earlier, could just put in a flag
	 */
	if(LM_STRING_CHR(addr, '@', index))
	{ 
		/* We didn't find an @ sign so we have local delivery */
		LM_STRING_DUP(user, addr);
		LM_STRING_NEW(domain, "");
	}
	/*  Seperate user and domain parts */
	else
	{
		/* We found an at, so build the user and domain strings */
		LM_STRING_DUP(user, addr);
		LM_STRING_DUP(domain, addr);

		/* Shift off the username and domain respectively */
		LM_STRING_END_SHIFT(user, LM_STRING_LEN(user)-index);
		LM_STRING_START_SHIFT(domain, index+1);
	}

	/* Check if this is a virtual domain user */
	retcode = msd_build_vdomain_user(&user, domain);
	is_vdomain_user = !retcode;

	/* Reset the matched a dash default file flag */
	/* This is set during one of the next two check_rcpt_.* 
	 * calls that follow.  FIXME 
	 */
	matched_dash_default = 0;

	/* Check if this user is in the qmail/users cdb file */
	retval = 0;
 	retcode = msd_check_rcpt_qmail_users(user, domain, &sysuser, &dotspam); 
	if(retcode > 0)
	{
		LM_ASSERT(dotspam == NULL);

		/* Check if this is a system user */
		retcode = msd_check_rcpt_system_user(user, domain, &sysuser, &dotspam);
		is_system_user = !retcode;
		if(retcode > 0)
		{
			LM_ASSERT(dotspam == NULL);
			/* Check if the username is an alias */
			retcode = msd_check_rcpt_alias(user, domain, &sysuser, &dotspam);
			is_alias = !retcode;
		}
	}

	/* The logic here, is as follows:
	 * if we received an address which is going to a virtual domain,
	 * the virtual domain will have a username prefix to add to the 
	 * user specified in the original address (prefix-user).  If this 
	 * address then matched a system user, and the address matched 
	 * a dot qmail default file (.qmail-vdomain-default or .qmail-default)
	 * then we have no way of knowing if the mail sent to this virtual domain
	 * is to a real address in the virtual domain.  To solve this, we have
	 * the user specify an external program which we will provide with the 
	 * original address and domain (user@domain), which will then do some
	 * magic (database lookup of some sort) and tell us if the address is
	 * valid for the domain.  If the program isn't specified, the user is 
	 * going to be considered valid.  If the program returns 0, the user 
	 * is valid.  If the program returns greater than zero, the user is
	 * invalid, and does not exist.  If the program returns less than zero, 
	 * an error occurred during the lookup, and the mail will be passed through
	 * to qmail since we don't want to lose any mail due to errors 
	 */
	if((retval == 0) && (is_vdomain_user != 0) &&
			(is_system_user != 0) && (matched_dash_default != 0))
	{
		retcode = msd_check_vdomain_user(user, domain);
	}

	if(retcode > 0)
	{
		/* User does not exist, blocking */
		retval = 1;
	}
	else if (retcode < 0)
	{
		/* There was some sort of error during validation
		   assume the user is valid and let qmail handle it */
		retval = 0;
	}
	else
	{
		retval = 0;
	}

	if(retcode == 0)
	{
		LM_ASSERT(sysuser != NULL);
		if(sysuser != NULL)
		{
			/* If the return code is zero, the user exists, and if dotspam is pointing
			   to something, they had a .magicspam file which matched this address */
		  SYSLOG((LOG_INFO, "DASHFILE used for checking user: %s", LM_STRING_BUFFER(dotspam)));
		}
	}

	/* we know now that the user exists and have the value of the system user 
	 * the mail address matched and a pointer to the .magicspam file that matched 
	 * the address if it was present 
	 */

	/* Now we check to see if the user is able to use spam checking as set 
	 * by the superuser
	 */
	if((retval == 0) && (config.per_user_spam_enable) == 1)
	{
		/* Check for the existance of the control file to enable spam checking on 
		 * this user's account */
		LM_ASSERT(sysuser != NULL);
		LM_ASSERT(dotspam != NULL);
		if(msd_user_spam_enabled(sysuser) != 0)
		{
		  /* Free the dotspam pointer to show that the user does not have
		   * spam checking enabled by the administrator */
		  SYSLOG((LOG_INFO, "User does not have spam checking enabled in %s/spam_users", LM_STRING_BUFFER(config.controldir)));
		  LM_STRING_FREE(dotspam);
		  retval = -1;
		}
	}

	/* Return the value of dotspam */
	LM_ASSERT(*pdotspam == NULL);
	*pdotspam = dotspam;

	/* Free up the strings we created */
	LM_STRING_FREE(user);
	LM_STRING_FREE(domain);
	if(sysuser != NULL) LM_STRING_FREE(sysuser);

	return retval;
}

/* Check to see if a user enable control file is present */
static int msd_user_spam_enabled(lm_string_t user)
{
	lm_string_t controlfile = NULL;
	struct stat buf;
	int retval = 0;

	if(LM_STRING_DUP(controlfile, spam_users_dir) ||
			LM_STRING_CAT(controlfile, user))
		return -1;

	/* Check for the existance of the file, any error means file doesn't exist */
	if(stat(LM_STRING_BUFFER(controlfile), &buf) != 0)
		retval = 1;

	LM_STRING_FREE(controlfile);
	return retval;
}

/* This routine will call the external program specified in the
 * configuration and pass it the username and domain arguments that
 * were originally recieved in the rcpt command.  The program is
 * expected to take the username and domain and return 0 if the user
 * is a valid user, greater than zero if it is not and less than zero
 * on any other error 
 */
static int msd_check_vdomain_user(lm_string_t user, lm_string_t domain)
{
	int retval, wpret;
	char childret;
	int childpid;
	char *prog;
	char *args[] = {
		NULL, NULL, NULL, NULL
	};

	/* Default return value for this function is user valid, as 
	 * If no program is specified, we already know we matched a -default
	 * file (or else this function wouldn't have been called).  This function
	 * just provides an extra level for virtual domain hosters to check a 
	 * database or whatever they want.
	 */
	retval = 0;

	/* Only do this if a virtual domain user checking program was specified */
	if(LM_STRING_LEN(config.check_vdomain_prog) > 0)
	{
		switch(childpid = fork())
		{
			case 0:
				/* Launch the program in the child process */
				prog = LM_STRING_BUFFER(config.check_vdomain_prog);
				args[0] = prog;
				args[1] = LM_STRING_BUFFER(user);
				args[2] = LM_STRING_BUFFER(domain);

			    if(execvp(prog, args))
			    {
			    	perror("vdomain user authentication program exec error!");
			    }
			    
			    perror("unreachable code");
			    _exit(-1);
			    break;
        
            case -1:
	      perror("forking vdomain auth program");
	      _exit(-1);
	      break;

	}

	/* Wait for the child to finish */
	do {
		wpret = waitpid(childpid, &retval, 0);
	} while ((wpret == -1) && (errno == EINTR));

	if(wpret != childpid)
	{
		/* Something bad happened */
		return -1;
	}

	/* Doing this to be safe.  WEXTISTATUS returns 0-255, which won't 
	 * do if our return value is an int.  Just be sure that it gets 
	 * sign extended by saving the exit status to a char and then 
	 * implicitly casting it to an int with the next assign 
	 */
	childret = WEXITSTATUS(retval);
	retval = childret;
	}

	return retval;
}

/* Attempt to find a dot dash file (.base-*) which
 * will be used to recieve mail for the given local user
 * to the given email address in the the provided
 * home directory 
 */
/* This function has been modified to take any file prefix and search
 * for dash files.  It was originally for only .qmail-* files but since
 * we modified the design for lmsmtpd to use the same format files, this
 * function proved to be valuable elsewhere 
 */
static int msd_find_qmail_dashfile_for_user(lm_string_t base, lm_string_t user, lm_string_t addr, lm_string_t homedir, lm_string_t *pdashfile)
{
	lm_string_t filename = NULL;
	lm_string_t exts = NULL;
	struct stat buf;
	int index;

	/* Strip the username we found off the given address username to
	 * leave only the extentions after the username 
	 */
	if(LM_STRING_NEW(exts, LM_STRING_BUFFER(addr)+LM_STRING_LEN(user)+1))
	{
		return -1;
	}

	/* Convert exts to lower case, convert '.' to ':' */
	LM_STRING_LOWER(exts);
	LM_STRING_REPLACE(exts, '.', ':');

	/* Check for ~user/.base-ext */
	if(LM_STRING_DUP(filename, homedir) ||
			LM_STRING_CAT(filename, slash) ||
			LM_STRING_CAT(filename, base) ||
			LM_STRING_CAT(filename, dash) ||
			LM_STRING_CAT(filename, exts))
	{
		return -1;
	}
	/* stat file and return sucess if it was there */
	if(!stat(LM_STRING_BUFFER(filename), &buf))
	{
		/* If the file is a directory put a / on the end */
		if(S_ISDIR(buf.st_mode) != 0)
		{
			if(LM_STRING_CAT(filename, slash) != 0)
				return -1;
		}

		if(exts != NULL) LM_STRING_FREE(exts);
		LM_ASSERT(*pdashfile == NULL);
		*pdashfile = filename;
		return 0;
	}
	else
	{
		/* Any error besides file not found makes us stop */
		if(errno != ENOENT)
		{
			if(exts != NULL) LM_STRING_FREE(exts);
			if(filename != NULL) LM_STRING_FREE(filename);
			return -1;
		}
	}

	LM_STRING_FREE(filename);

	/* If we have a prefix-ext type extension, we have to recurse
	 * backwards through the extention and put a -default on it and 
	 * check for the file.
	 */
	index = 0;

	while(index != -1)
	{
		lm_string_t prefix = NULL;

		/* Check to see if we have a prefix-ext type extension */
		if(!LM_STRING_RCHR(exts, '-', index))
		{
			/* Shift off the extension */
			LM_STRING_END_SHIFT(exts, LM_STRING_LEN(exts)-index);
		}
		else
		{
			continue;
		}


		/* Check for ~user/.prefix-prefix-default */
		if(LM_STRING_DUP(prefix, exts) ||
				LM_STRING_DUP(filename, homedir) ||
				LM_STRING_CAT(filename, slash) || 
				LM_STRING_CAT(filename, base) ||
				LM_STRING_CAT(filename, dash) ||
				LM_STRING_CAT(filename, prefix) ||
				LM_STRING_CAT(filename, dashdefault))
		{
			return -1;
		}
		/* stat file and return sucess if it was there */
		if(!stat(LM_STRING_BUFFER(filename), &buf))
		{
			/* If the file is a directory put a / on the end */
			if(S_ISDIR(buf.st_mode) != 0)
			{
				if(LM_STRING_CAT(filename, slash) != 0)
					return -1;
			}

			LM_STRING_FREE(prefix);
			if(exts != NULL) LM_STRING_FREE(exts);
			matched_dash_default = 1;
			LM_ASSERT(*pdashfile == NULL);
			*pdashfile = filename;
			return 0;
		}
		else
		{
			/* Any error besides file not found makes us stop */
			if(errno != ENOENT)
			{
				if(exts != NULL) LM_STRING_FREE(exts);
				if(filename != NULL) LM_STRING_FREE(filename);
				return -1;
			}
		}

		LM_STRING_FREE(prefix);
		LM_STRING_FREE(filename);
	}

	/* Check for ~user/.qmail-default */
	if(LM_STRING_DUP(filename, homedir) ||
			LM_STRING_CAT(filename, slash) ||
			LM_STRING_CAT(filename, base) ||
			LM_STRING_CAT(filename, dashdefault))
	{
		return -1;
	}

	if(!stat(LM_STRING_BUFFER(filename), &buf))
	{
		/* The file existed so the user is valid */
		/* If the file is a directory put a / on the end */
		if(S_ISDIR(buf.st_mode) != 0)
		{
			if(LM_STRING_CAT(filename, slash) != 0)
				return -1;
		}

		if(exts != NULL) LM_STRING_FREE(exts);
		LM_ASSERT(*pdashfile == NULL);
		*pdashfile = filename;
		matched_dash_default = 1;
		return 0;
	}
	else
	{
		/* Any error besides file not found makes us stop */
		if(errno != ENOENT)
		{
			if(exts != NULL) LM_STRING_FREE(exts);
			if(filename != NULL) LM_STRING_FREE(filename);
			return -1;
		}
	}
	LM_STRING_FREE(filename);

	if(exts != NULL) LM_STRING_FREE(exts);
	if(filename != NULL) LM_STRING_FREE(filename);
	return 1;
}

/* Attempt to find the given user address in the qmail user's file
 * and then try and find a dot qmail file which will match that
 * address if the message was to be delivered.
 */
static int msd_check_rcpt_qmail_users(lm_string_t user, lm_string_t domain, lm_string_t* psysuser, lm_string_t* pdashfile)
{
	int has_dash;
	lm_string_t cdb_user = NULL;
	int cdb_user_len;
	int retval = 1;
	lm_string_t username = NULL;
	lm_string_t homedir = NULL;
	lm_string_t dashfile = NULL;

	/* Check if the users file is open, return does not exist if not */
	if(cdbusers == NULL)
		return 1;

	/* See if the username has a dash in it */
	if(!LM_STRING_CHR(user, '-', has_dash))
	{
		/* Check for wildcard in cdb file */
		if(LM_STRING_DUP(cdb_user, bang) || 
				LM_STRING_CAT(cdb_user, user))
		{
			return -1;
		}
		LM_STRING_END_SHIFT(cdb_user, LM_STRING_LEN(user)-has_dash-1);
		cdb_user_len = LM_STRING_LEN(cdb_user);

		/* Build the username from the cdb search name */
		if(LM_STRING_DUP(username, cdb_user))
		{
			LM_STRING_FREE(cdb_user);
			return -1;
		}
		LM_STRING_START_SHIFT(username, 1);
		LM_STRING_END_SHIFT(username, 1);
	}
	else
	{
		/* Check for non-wildcard user assignment in cdb file */
		LM_STRING_DUP(cdb_user, bang);
		LM_STRING_CAT(cdb_user, user);
		cdb_user_len = LM_STRING_LEN(cdb_user)+1;

		/* Build the username from the cdb search name */
		LM_STRING_DUP(username, cdb_user);
		LM_STRING_START_SHIFT(username, 1);
	}


	/* Attempt to find the user in the cdb file */
	retval = cdb_lookup(cdbusers, LM_STRING_BUFFER(cdb_user), cdb_user_len, &homedir);

	/* If we had a wildcard match, we need to search for the appropriate
	 * dash file to verify that the user exists
	 */
	/* If we had a wildcard match, we need to search for the appropriate
	 * dash file to verify that the user exists
	 */
	if(retval == 0)
	{
		if(has_dash > 0)
		{
			/* try and find a dot qmail file which will accept mail to this address */
			retval = msd_find_qmail_dashfile_for_user(dotqmail, username, user, homedir, &dashfile);

			/* search for a .magicspam file that matches the address as well */
			if(retval == 0)
			{
				int temp;
				lm_string_t tmpstring = NULL;

				LM_ASSERT(dashfile);
				LM_STRING_FREE(dashfile);

				/* Find the matching .magicspam file for the valid user */
				temp = msd_find_qmail_dashfile_for_user(dotmagicspam, username, user, homedir, &dashfile);
				LM_ASSERT(*pdashfile == NULL);
				*pdashfile = dashfile;
				LM_ASSERT(*psysuser == NULL);
				if(LM_STRING_DUP(tmpstring, username))
					return -1;
				*psysuser = tmpstring;
			}
		}
		else
		{
			struct stat buf;
			lm_string_t tmpstring = NULL;

			if(LM_STRING_DUP(dashfile, homedir) ||
					LM_STRING_CAT(dashfile, slash) ||
					LM_STRING_CAT(dashfile, dotmagicspam))
				return -1;
			/* Check to see if the dashfile name is a directory
			 * and put a trailing / on it if it is */
			if(stat(LM_STRING_BUFFER(dashfile), &buf) == 0)
				if(S_ISDIR(buf.st_mode) != 0)
					if(LM_STRING_CAT(dashfile, slash) != 0)
						return -1;
			LM_ASSERT(*pdashfile == NULL);
			*pdashfile = dashfile;
			LM_ASSERT(*psysuser == NULL);
			if(LM_STRING_DUP(tmpstring, username))
				return -1;
			*psysuser = tmpstring;
			retval = 0;
		}
	}

	/* Free up memory */
	if(homedir != NULL) LM_STRING_FREE(homedir);
	if(username != NULL) LM_STRING_FREE(username);
	if(cdb_user != NULL) LM_STRING_FREE(cdb_user);

	return retval;
}

/* Attempt to verify that the given user exists by checking
 * the /etc/passwd file
 */
static int msd_check_rcpt_system_user(lm_string_t user, lm_string_t domain, lm_string_t* psysuser, lm_string_t* pdashfile)
{
	struct passwd *pw;
	int index;
	int has_dash = 0;
	int retval = 1;
	lm_string_t tmpuser = NULL;
	lm_string_t dashfile = NULL;

	/* Make a copy of user so we can trash it */
	if(LM_STRING_DUP(tmpuser, user))
		return -1;

	/* Check for valid system users */
	while(LM_STRING_LEN(tmpuser) > 0)
	{
		pw = getpwnam(LM_STRING_BUFFER(tmpuser));
		if(pw != NULL)
		{
			lm_string_t homedir = NULL;

			/* We found a valid user */
			if(has_dash == 0)
			{
				struct stat buf;

				/* we need to do a bit more checking here to see if 
				 * the user is valid.  Qmail checks to see if the home directory
				 * returned in the pw structure is owned by the uid that is returned.
				 * that might be the best thing to do 
				 */
				/* If we have a home directory */
				if(pw->pw_dir[0] != '\0')
				{
					/* Stat the file to get it's info */
					if(stat(pw->pw_dir, &buf))
					{
						LM_STRING_FREE(tmpuser);
						return -1;
					}
					/* If the uid and gid of the user's home directory match
					 * thier uid and gid, let's assume the user is valid and
					 * not a pseudo user 
					 */
					if((pw->pw_uid == buf.st_uid) &&
							(pw->pw_gid == buf.st_gid))
					{
						lm_string_t tmpstring = NULL;

						/* Build the .magicspam filename that matches this user address*/
						if(LM_STRING_NEW(homedir, pw->pw_dir) ||
								LM_STRING_CAT(homedir, slash) ||
								LM_STRING_CAT(homedir, dotmagicspam))
							return -1;

						/* Put a trailing / on the filename if it is a directory */
						if(stat(LM_STRING_BUFFER(homedir), &buf) == 0)
						{
							if(S_ISDIR(buf.st_mode) != 0)
								if(LM_STRING_CAT(homedir, slash))
									return -1;
						}

						/* Assign the return values */
						LM_ASSERT(*pdashfile == NULL);
						*pdashfile = homedir;
						LM_ASSERT(*psysuser == NULL);
						if(LM_STRING_DUP(tmpstring, tmpuser))
							return -1;
						*psysuser = tmpstring;
						retval = 0;
					}
					else
					{
						retval = 1;
					}

				}
			}
			else
			{
				/* The username has a dash in it, and we might have stripped off 
				 * some of it to find the user that we currently have a password entry 
				 * for.  We now have to look for a .qmail type file which will match 
				 * the given addres, since the mail will bounce if one is not present, 
				 * and we want to minimize bounces to non-existant adresses.
				 */
				if(LM_STRING_NEW(homedir, pw->pw_dir))
					return -1;
				retval = msd_find_qmail_dashfile_for_user(dotqmail, tmpuser, user, homedir, &dashfile);

				/* If we found a user, return the value of homedir, else free it */
				LM_ASSERT(*pdashfile == NULL);
				if(retval == 0)
				{
					int temp;
					lm_string_t tmpstring = NULL;

					LM_ASSERT(dashfile);
					LM_STRING_FREE(dashfile);

					/* Find the matching .magicspam file for the valid user */
					temp = msd_find_qmail_dashfile_for_user(dotmagicspam, tmpuser, user, homedir, &dashfile);
					LM_ASSERT(*pdashfile == NULL);
					*pdashfile = dashfile;
					LM_ASSERT(*psysuser == NULL);
					if(LM_STRING_DUP(tmpstring, tmpuser))
						return -1;
					*psysuser = tmpstring;
				}

				LM_STRING_FREE(homedir);
			}

			/* Get out of the loop since we found a valid user */
			break;
		}
		else
		{
			/* 
			 * Look for a dash in the user name, find the last one and
			 * shift the part after the dash and the dash off the end 
			 * and then look again 
			 */
			if(!LM_STRING_RCHR(tmpuser, '-', index))
			{
				/* Shift off the last extention argument and continue with the loop */
				has_dash = 1;
				LM_STRING_END_SHIFT(tmpuser, LM_STRING_LEN(tmpuser)-index);
			}
			else
			{
				/* No dash, just shift off the whole string */
				LM_STRING_END_SHIFT(tmpuser, LM_STRING_LEN(tmpuser));
				LM_ASSERT(LM_STRING_LEN(tmpuser) == 0);
			}
		}
	}

	LM_STRING_FREE(tmpuser);

	return retval;
}

/* Search the vdomains file and see if the given domain is a virtual domain.
 * If it is, get the prefix value specified for the domain and put it on
 * the username.
 */
static int msd_build_vdomain_user(lm_string_t *puser, lm_string_t domain)
{
	lm_string_t virtualdomains = NULL;
	lm_string_t vdomain = NULL;
	lm_string_t vprefix = NULL;
	lm_string_t user;
	lm_string_t newuser = NULL;
	int index= -1;
	int match = 0;


	LM_ASSERT(*puser != NULL);
	user = *puser;

	/* Check if domain is in virtualdomains */
	if(LM_STRING_DUP(virtualdomains, config_qmail.virtualdomains))
	{
		return -1;
	}

	/* Go through each virtual domain and see if there is a match */
	while((user != newuser) && 
			(LM_STRING_LEN(virtualdomains) > 0))
	{
		if(LM_STRING_GET_LINE(virtualdomains, vdomain))
		{
			LM_STRING_FREE(virtualdomains);
			return -1;
		}

		if(LM_STRING_CHR(vdomain, ':', index))
		{
			/* We couldn't find a :, error with file format */
			LM_STRING_FREE(virtualdomains);
			LM_STRING_FREE(vdomain);
			return -1;
		}

		/* check for catchall alias (see man qmail-send virtualdomains) */
		if(index == 0)
		{
			lm_string_t tempvdomain = NULL;

			/* We got the catchall alias for all domains, so let's make up
			 * a fake record with the current domain as the virtual domain, with
			 * the same catchall prepend 
			 */
			if(LM_STRING_NEW(tempvdomain, LM_STRING_BUFFER(domain)) ||
					LM_STRING_CAT(tempvdomain, vdomain))
				return -1;
			LM_STRING_FREE(vdomain);
			vdomain = tempvdomain;

			/* Point index to the colon in the  new string */
			index = LM_STRING_LEN(domain);
		}

		/* Check for an empty prepend, which means that the 
		 * domain listed is not a virtual domain
		 */
		if(LM_STRING_LEN(vdomain)+1 == index)
		{
			/* From man qmail-send:
			 * virtualdomains may also contain exceptions: an empty prepend
			 * means that domain is not a virtual domain.
			 * Let's just skip over these and continue looking through the list.
			 */
			/* TODO I'm not quite sure if this is the correct way to handle this.  
			 * Maybe we should be compairing the domain to this one and returning a
			 * user does not exist?  I don't think continuing through the rest of the
			 * list after skipping this one will do much harm, but I could be wrong.
			 */
			LM_STRING_FREE(vdomain);
			continue;
		}

		/* Make a copy of the domain so we can get both the domain and the prefix */
		if(LM_STRING_DUP(vprefix, vdomain))
		{
			/* Allocation error */
			LM_STRING_FREE(virtualdomains);
			LM_STRING_FREE(vdomain);
			return -1;
		}

		/* Build the prefix and domain strings */
		LM_STRING_START_SHIFT(vprefix, index+1);
		LM_STRING_END_SHIFT(vdomain, LM_STRING_LEN(vdomain)-index);

		/* Check for period character at the start of the virtual 
		 * domain, which indicates a wildcard.
		 */
		match = smtp_match_domain(domain, vdomain);

		/* Check for match with current line */
		if(!match)
		{
			/* We did find a match */
			/* Create prefix-user and check for system user */
			if(LM_STRING_DUP(newuser, vprefix) ||
					LM_STRING_CAT(newuser, dash) ||
					LM_STRING_CAT(newuser, user))
			{
				LM_STRING_FREE(virtualdomains);
				LM_STRING_FREE(vprefix);
				LM_STRING_FREE(vdomain);
				return -1;
			}

			LM_STRING_FREE(user);
			is_vdomain_addr = 1;
			user = newuser;
		}
		LM_STRING_FREE(vprefix);
		LM_STRING_FREE(vdomain);
	}
	LM_STRING_FREE(virtualdomains);

	/* Return non valid user for now */
	*puser = user;
	return (user != newuser) ? 1 : 0;
}

/* Check to see if the given user is a qmail alias.  Qmail aliases are 
 * found in the home directory of the alias user, which is found in either 
 * the password file or the cdb file.  For an alias to exist, a 
 * ~alias/.qmail-<alias> must exist, so that is what we check for.  We can
 * then assume that the user does exist since an alias file can *only* 
 * contain the name of a system user, which will always exist (and if it 
 * doesn't, that's not our problem).
 */
static int msd_check_rcpt_alias(lm_string_t user, lm_string_t domain, lm_string_t* psysuser, lm_string_t *pdashfile)
{
	lm_string_t filename = NULL;
	lm_string_t dashfile = NULL;
	lm_string_t tmpuser = NULL;
	int retval = 1;
	struct stat buf;


	/* Build ~alias/.qmail-user */
	LM_ASSERT(qmailalias);

	if(LM_STRING_DUP(filename, qmailalias) ||
			LM_STRING_CAT(filename,user)) 
		return -1;

	/* Check to see if the file ~alias/.qmail-user exists */
	if(!stat(LM_STRING_BUFFER(filename), &buf))
	{
		int temp;
		lm_string_t tmpstring = NULL;

		/* build the alias user name for finding the dot magicspam file * */
		if(LM_STRING_DUP(tmpuser, alias) ||
				LM_STRING_CAT(tmpuser, dash) ||
				LM_STRING_CAT(tmpuser, user))
		{
			return -1;
		}

		/* Find the matching .magicspam file for the valid user */
		temp = msd_find_qmail_dashfile_for_user(dotmagicspam, alias, tmpuser, qmailaliashome, &dashfile);
		LM_ASSERT(*pdashfile == NULL);
		*pdashfile = dashfile;
		LM_ASSERT(*psysuser == NULL);
		if(LM_STRING_DUP(tmpstring, alias))
			return -1;
		*psysuser = tmpstring;

		/* Clean up allocated strings */
		LM_STRING_FREE(tmpuser);
		LM_STRING_FREE(filename);
		return 0;
	}
	else
	{
		if(errno != ENOENT)
			retval = -1;
		else
			retval = 1;
	}

	LM_STRING_FREE(filename);

	/* Return valid user for now */
	return retval;
}
