
/*
 * COPYRIGHT INFORMATION - DO NOT REMOVE
 * "Portions Copyright (c) 2000-2001 LinuxMagic Inc. All Rights Reserved.
 *
 * This file contains Original Code and/or Modifications of Original Code as
 * defined in and that are subject to the Free Source Code License Version
 * 1.0 (the 'License'). You may not use this file except in compliance with
 * the License. Please obtain a copy of the License at:
 *
 * http://www.linuxmagic.com/opensource/licensing/
 *
 * and read it before using this file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND LINUXMAGIC HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT. Please see
 * the License for the specific language governing rights and limitations
 * under the License."
 *
 * Please read the terms of this license carefully. By using or downloading
 * this software or file, you are accepting and agreeing to the terms of this
 * license with LinuxMagic Inc. If you are agreeing to this license on behalf
 * of a company, you represent that you are authorized to bind the company to
 * such a license. If you do not meet this criterion or you do not agree to
 * any of the terms of this license, do NOT download, distribute, use or alter
 * this software or file in any way.
 *
 * Author: Burton Samograd (burton@wizard.ca)
 * CVS Id: $Id: smtp_check.c,v 1.7 2003/01/15 18:25:47 burton Exp $
 *
 * DO NOT MODIFY WITHOUT CONSULTING THE LICENSE
 */


/* C includes */
#include <stdlib.h>
#include <string.h>

/* System includes */
#include <unistd.h>
#include <netdb.h>


#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>

/* lmlibs */
#include <liblm.h>

/* magic-smtpd includes */
#include "smtp_check.h"
#include "config_msd.h"
#include "utils.h"

int require_helo(lm_string_t helostring)
{
  int retval = 0;
  char *buffer;

  buffer = LM_STRING_BUFFER(helostring);

  if(buffer[0] == 'u' &&
     buffer[1] == 'n' &&
     buffer[2] == 'k' &&
     buffer[3] == 'n' &&
     buffer[4] == 'o' &&
     buffer[5] == 'w' &&
     buffer[6] == 'n')
    retval = -1;
  return retval;
}

int smtp_check_valid_domain(lm_string_t domain)
{
  char buffer[1024];
  int retval = 0;

  /* If given an ip address don't bother resolving and say it's valid */
  if(!is_ip(domain))
    {
      SYSLOG((LOG_INFO, "Did not attempt to resolve IP address"));
      return 0;
    }
  
  SYSLOG((LOG_INFO, "Lookup up domain %s (this may take a while)", LM_STRING_BUFFER(domain)));

  retval = res_query(LM_STRING_BUFFER(domain), C_IN, T_A, buffer, sizeof(buffer));
  if(retval == -1)
    retval = res_query(LM_STRING_BUFFER(domain), C_IN, T_MX, buffer, sizeof(buffer));
  return retval == -1;
}

int valid_helo_domain(lm_string_t domain)
{
  /* default fail return value since zero length domain wont resolve */
  int retval = -1;
  
  if(LM_STRING_LEN(domain) > 0)
    {
      retval = smtp_check_valid_domain(domain);
    }
  return retval;
}

int valid_from_domain(lm_string_t domain)
{
  /* default fail return value since zero length domain wont resolve */
  int retval = -1;
  if(LM_STRING_LEN(domain) > 0)
    {
      retval = smtp_check_valid_domain(domain);
    }
  return retval;
}

int block_mail_from_self(lm_string_t from, lm_string_t to)
{
  int retval;
  
  /* Very basic check to compare the two addresses */
  retval = !LM_STRING_CMP(from, to);
  return retval;
}

int require_full_addr(lm_string_t addr)
{
  int retval = 1;
  int index;

  /* If we have an @ sign and it wasn't
     the first or last character*/
  if(!LM_STRING_CHR(addr, '@', index) &&
     (index != 0) &&
     LM_STRING_LEN(addr) != (index -1))
    {
      retval = 0;
    }
  return retval;
}

int smtp_blocking(void)
{
  return 0;
}

int valid_bounce(lm_string_t fromdomain)
{
  int retval = 0;
  char buffer[1024];

  memset(buffer, 0, sizeof(buffer));
  
  retval = res_query(LM_STRING_BUFFER(fromdomain), C_IN, T_MX, buffer, sizeof(buffer));
  
  return retval < 0;
}

int check_ip_reverse_dns(lm_string_t ip)
{
  struct hostent *host = NULL;
  int retval = 0;

  /* If this is not a d.d.d.d type string don't lookup */
  if(is_ip(ip))
    return 0;
  
  /* gethostbyaddr */
  while((host == gethostbyaddr(LM_STRING_BUFFER(ip),
			      LM_STRING_LEN(ip), AF_INET)) &&
	(h_errno == TRY_AGAIN))
    ; /* do nothing */
  
  /* Check for error */
  if(host == NULL)
    {
      switch(h_errno)
	{
	case HOST_NOT_FOUND:
	  SYSLOG((LOG_INFO, "Hostname for IP %s could not be found", LM_STRING_BUFFER(ip)));
	  retval = 1;
	  break;
	case NO_RECOVERY:
	  SYSLOG((LOG_INFO, "A non-recoverable DNS server error occured."));
	  retval = -1;
	  break;
	default:
	  /* everything ok, return sucess */
	  retval = 0;
	}
    }
  return retval;
}
